#include "localsolver.h"
#include <fstream>
#include <iostream>
#include <sstream>
#include <vector>

using namespace localsolver;
using namespace std;

class Facilitylocation {
public:
    // Number of locations
    int N;

    // Size of the subset S of facilities
    int p;

    // Weight matrix of the shortest path between locations
    vector<vector<int>> w;

    // Maximum distance between two locations
    int wmax;

    // LocalSolver
    LocalSolver localsolver;

    // Decisions variables
    vector<LSExpression> x;

    // Objective
    LSExpression totalCost;

    // Read instance data
    void readInstance(const string& fileName) {
        ifstream infile;
        infile.exceptions(ifstream::failbit | ifstream::badbit);
        infile.open(fileName.c_str());

        int tmp;
        infile >> N;
        infile >> tmp; // Skip number of edges
        infile >> p;

        w.resize(N);
        wmax = 0;
        for (int i = 0; i < N; ++i) {
            w[i].resize(N);
            for (int j = 0; j < N; ++j) {
                infile >> w[i][j];
                if (w[i][j] > wmax) {
                    wmax = w[i][j];
                }
            }
        }
    }

    // Declare the optimization model
    void solve(int limit) {
        LSModel m = localsolver.getModel();

        // One variable for each location: 1 if facility, 0 otherwise
        x.resize(N);
        for (int i = 0; i < N; ++i)
            x[i] = m.boolVar();

        // No more than p locations are selected to be facilities
        LSExpression openedLocations = m.sum(x.begin(), x.end());
        m.constraint(openedLocations <= p);

        // Costs between location i and j is w[i][j] if j is a facility or 2 * wmax if not
        vector<vector<LSExpression>> costs(N);
        for (int i = 0; i < N; ++i) {
            costs[i].resize(N);
            for (int j = 0; j < N; ++j) {
                costs[i][j] = m.iif(x[j], w[i][j], 2 * wmax);
            }
        }

        // Cost between location i and the closest facility
        vector<LSExpression> cost(N);
        for (int i = 0; i < N; ++i) {
            cost[i] = m.min(costs[i].begin(), costs[i].end());
        }

        // Minimize the total cost
        totalCost = m.sum(cost.begin(), cost.end());
        m.minimize(totalCost);
        m.close();

        // Parameterize the solver
        localsolver.getParam().setTimeLimit(limit);

        localsolver.solve();
    }

    /* Write the solution in a file with the following format: 
     * - value of the objective
     * - indices of the facilities (between 0 and N-1) */
    void writeSolution(const string& fileName) {
        ofstream outfile;
        outfile.exceptions(ofstream::failbit | ofstream::badbit);
        outfile.open(fileName.c_str());

        outfile << totalCost.getValue() << endl;
        for (int i = 0; i < N; ++i) {
            if (x[i].getValue() == 1)
                outfile << i << " ";
        }
        outfile << endl;
    }
};

int main(int argc, char** argv) {
    if (argc < 2) {
        cerr << "Usage: facility_location inputFile [outputFile] [timeLimit] " << endl;
        return 1;
    }

    const char* instanceFile = argv[1];
    const char* solFile = argc > 2 ? argv[2] : NULL;
    const char* strTimeLimit = argc > 3 ? argv[3] : "20";

    try {
        Facilitylocation model;
        model.readInstance(instanceFile);
        model.solve(atoi(strTimeLimit));
        if (solFile != NULL)
            model.writeSolution(solFile);
        return 0;
    } catch (const exception& e) {
        cerr << "An error occurred: " << e.what() << endl;
        return 1;
    }
}
